#!/bin/sh
#
# Core functions and main program entry point.
#
# Copyright 2025 Andrew Wood
#
# License GPLv3+: GNU GPL version 3 or later; see `docs/COPYING'.
#

# Adjust the default path.
# Note that /usr/local is vital on BSDs for tools like gpg.
export PATH="/sbin:/bin:/usr/sbin:/usr/bin:/usr/local/bin:/usr/local/sbin:$PATH"

# Explicitly set the umask.
umask 0022

#
# Core functions.
#

# Output the message $1 to standard error, prefixed with the program name
# and a colon.
#
reportError () {
	printf '%s: %s\n' "${PROGRAM_NAME}" "$1" 1>&2
}

# Call reportError and provide an additional reminder about how to check the
# program syntax.
#
reportArgumentsError () {
	reportError "$*"
	printf '%s\n' "Try \`${PROGRAM_NAME} --help' for more information." 1>&2
}

# Output the path of this script.
#
locateSelf () {
	if test -e "${bindir}/${PROGRAM_NAME}"; then
		printf '%s\n' "${bindir}/${PROGRAM_NAME}"
		return 0
	elif test -e "${srcdir}/${PROGRAM_NAME}"; then
		printf '%s\n' "${srcdir}/${PROGRAM_NAME}"
		return 0
	elif test -e "$0"; then
		printf '%s\n' "$0"
		return 0
	fi
	reportError 'cannot locate path to this script'
	return "${RC_LOCAL_FAULT}"
}

# Output the path of component $1.
#
locateComponent () {
	if test -e "${componentDir}/$1.sh"; then
		printf '%s\n' "${componentDir}/$1.sh"
		return 0
	elif test -e "${srcdir}/src/$1.sh"; then
		printf '%s\n' "${srcdir}/src/$1.sh"
		return 0
	fi
	reportError "$1: component not found"
	return "${RC_LOCAL_FAULT}"
}

# Load the component $1.
#
loadComponent () {
	componentPath="$(locateComponent "$1")" || return $?
	. "${componentPath}"
	return $?
}

# Load the appropriate component whose name starts with $1 for the current
# ${targetOs}, first looking for an exact match, and if there isn't one,
# falling back to a match of just the OS name with no version number.  Once
# the component is loaded, sets ${osSpecificComponent} to the name of the
# component that was loaded, and ${osSpecificFunction} to the component name
# with "-" replaced with "_".
#
loadOsSpecificComponent () {
	componentBase="$1"

	# First attempt using full targetOs.
	osSpecificComponent="${componentBase}-${targetOs}"
	osSpecificFunction="$(printf '%s\n' "${osSpecificComponent}" | tr '-' '_')"
	# Do nothing if it's already loaded.
	type "${osSpecificFunction}" >/dev/null 2>&1 && return 0
	loadComponent "${osSpecificComponent}" 2>/dev/null && return 0

	# Now try with the numbers removed from the end of targetOs.
	osSpecificComponent="$(printf '%s\n' "${osSpecificComponent}" | sed 's/[0-9]*$//')"
	osSpecificFunction="$(printf '%s\n' "${osSpecificComponent}" | tr '-' '_')"
	# Do nothing if it's already loaded.
	type "${osSpecificFunction}" >/dev/null 2>&1 && return 0

	loadComponent "${osSpecificComponent}"
	return $?
}

# Run the function $1 with the remaining parameters, unless function $1
# doesn't exist, in which case any numbers at the end of its name are
# removed and the resultant function is called instead.
#
runFunctionWithGenericFallback () {
	funcName="$1"
	shift
	type "${funcName}" >/dev/null 2>&1 || funcName="$(printf '%s\n' "${funcName}" | sed 's/[0-9]*$//')"
	"${funcName}" "$@"
	return $?
}

# Extract the source archive ${sourcePath} to the directory $1.  If the top
# level of the archive contains just one directory, that directory's
# contents are copied up one level, so for instance "$1/foo-1.2.3/bar" gets
# copied to "$1/bar".  The archive itself is also copied verbatim.
#
extractSource () {
	sourceExtractionDir="$1"

	case "${sourcePath}" in
	*.tar.gz|*.tgz) tar xzf "${sourcePath}" -C "${sourceExtractionDir}" || return "${RC_LOCAL_FAULT}" ;;
	*.tar.bz2|*.tbz2) tar xjf "${sourcePath}" -C "${sourceExtractionDir}" || return "${RC_LOCAL_FAULT}" ;;
	*.tar.xz|*.txz) tar xJf "${sourcePath}" -C "${sourceExtractionDir}" || return "${RC_LOCAL_FAULT}" ;;
	*.zip) unzip -d "${sourceExtractionDir}" "${sourcePath}" || return "${RC_LOCAL_FAULT}" ;;
	*) reportError "${sourcePath}: unknown file extension"; return "${RC_BAD_ARGS}" ;;
	esac

	# If the archive contained just one thing at the top level, assume
	# it's a directory, and copy the directory's contents up one level.
	if find "${sourceExtractionDir}" -mindepth 1 -maxdepth 1 2>/dev/null | wc -l | grep -Fqx '1'; then
		cp -al "${sourceExtractionDir}"/*/* "${sourceExtractionDir}"/
	fi

	# Copy the archive itself as well, since some build processes need
	# the original archive, not the extracted contents.
	cp "${sourcePath}" "${sourceExtractionDir}/" || return "${RC_LOCAL_FAULT}"

	return 0
}

# If ${workDir}/signing-key exists, import it into a temporary GnuPG
# keyring, setting GNUPGHOME to a temporary directory for the purpose, and
# setting $signingKey to the ID of the signing key.
#
setupSigningKey () {
	test -s "${workDir}/signing-key" || return 0

	if ! test "${action}" = 'inside-container'; then
		GNUPGHOME="${workDir}/gnupghome"
		mkdir -p "${GNUPGHOME}"
		chmod 700 "${GNUPGHOME}"
		export GNUPGHOME
	fi

	gpg --import "${workDir}/signing-key" || return "${RC_LOCAL_FAULT}"
	signingKey="$(gpg --list-keys --with-colons | awk -F : '$1=="pub" {print $5}')"

	return 0
}

# Check that $1 is a supported target operating system, exiting immediately
# if not.  Set targetOs to $1, and set targetOsFunctionName to $1 after
# mapping it to lower case and replacing "-" with "_".  Set buildImageName
# to ${imagePrefix}${targetOsFunctionName}, which gives the name to use for
# the container image for this OS.
#
setTargetOs () {
	targetOs="$1"
	printf '%s\n' "${supportedTargets}" | tr ' ' '\n' | grep -E . | grep -Fqx "${targetOs}" \
	|| { reportArgumentsError "${targetOs}: unknown target operating system"; exit "${RC_BAD_ARGS}"; }
	targetOsFunctionName="$(printf '%s\n' "${targetOs}" | tr '-' '_' | tr '[:upper:]' '[:lower:]')"
	buildImageName="${imagePrefix}${targetOsFunctionName}"
}

# Index a repository for target operating system $1, whose packages are in
# directory $2, and write the new index to directory $3.  Loads and calls
# the appropriate repository indexing function for the target OS.
#
indexRepository () {
	repoTargetOs="$1"
	repoSourceDir="$2"
	repoIndexDir="$3"
	repoCheckArg="$4"

	prevTargetOs="${targetOs}"
	setTargetOs "${repoTargetOs}"

	# Make sure the build image is present, if using containers.
	if ! "${nativeOnly}"; then
		loadComponent 'containers' || exit $?
		buildImage || exit $?
	fi

	if loadOsSpecificComponent 'index-repository'; then
		"${osSpecificFunction}" "${repoSourceDir}" "${repoIndexDir}" "${repoCheckArg}"
		indexResult="$?"
	else
		indexResult="${RC_LOCAL_FAULT}"
	fi

	test -n "${prevTargetOs}" && setTargetOs "${prevTargetOs}"
	return "${indexResult}"
}

# Add constraint $1 to the list of active constraints, exiting with an error
# if its syntax is incorrect.
#
addConstraint () {
	constraintSpec="$1"
	constraintKey="${constraintSpec%%=*}"
	constraintValue="${constraintSpec#*=}"
	case "${constraintKey}" in
	'')
		reportArgumentsError '--constraint: missing constraint'
		return "${RC_BAD_ARGS}"
		;;
	'pinned'|'included'|'outdated')
		case "$(printf '%s\n' "${constraintValue}" | tr '[:upper:]' '[:lower:]')" in
		'true'|'on'|'yes'|'1') constraintValue="true" ;;
		'false'|'off'|'no'|'0') constraintValue="false" ;;
		*)
			reportArgumentsError "--constraint: ${constraintKey}: ${constraintValue}: boolean value expected"
			return "${RC_BAD_ARGS}"
			;;
		esac
		;;
	'package'|'repository'|'target')
		if ! test -n "${constraintValue}"; then
			reportArgumentsError "--constraint: ${constraintKey}: missing value"
			return "${RC_BAD_ARGS}"
		fi
		;;
	*)
		reportArgumentsError "--constraint: ${constraintKey}: unknown constraint key"
		return "${RC_BAD_ARGS}"
		;;
	esac

	if test -n "${constraints}"; then
		constraints="$(printf '%s\n%s\n' "${constraints}" "${constraintKey} ${constraintValue}")"
	else
		constraints="${constraintKey} ${constraintValue}"
	fi
	return 0
}


#
# Variable initialisation and command-line option processing.
#

# Initial values for command-line option processing.
action=""
nativeOnly=false
targetOs=""
targetOsFunctionName=""
buildImageName=""
sourcePath=""
instructionsPath=""
destinationPath=""
keyFile=""
constraints=""
logUser="${LOGNAME}"
minArgs=0
maxArgs=0

# Parse the command-line options to apply options and select the action.
while test -n "$1" && test -z "${action}"; do
	arg="$1"
	shift
	case "${arg}" in
	'-h'|'--help'|'help') action='help' ;;
	'-V'|'--version'|'version') action='version' ;;
	'-n'|'--native') nativeOnly=true ;;
	'-t'|'--target') targetOs="$1"; shift 2>/dev/null ;;
	'--target='*) targetOs="${arg#*=}" ;;
	'-s'|'--source') sourcePath="$1"; shift 2>/dev/null ;;
	'--source='*) sourcePath="${arg#*=}" ;;
	'-i'|'--instructions') instructionsPath="$1"; shift 2>/dev/null ;;
	'--instructions='*) instructionsPath="${arg#*=}" ;;
	'-d'|'--destination') destinationPath="$1"; shift 2>/dev/null ;;
	'--destination='*) destinationPath="${arg#*=}" ;;
	'-k'|'--key') keyFile="$1"; shift 2>/dev/null ;;
	'--key='*) keyFile="${arg#*=}" ;;
	'-c'|'--constraint') addConstraint "$1" || exit $?; shift 2>/dev/null ;;
	'--constraint='*) addConstraint "${arg#*=}" || exit $? ;;
	'-u'|'--user') logUser="$1"; shift 2>/dev/null ;;
	'--user='*) logUser="${arg#*=}" ;;
	'-'*) reportArgumentsError "${arg}: unknown option"; exit "${RC_BAD_ARGS}" ;;
	'build-package') action="${arg}"; minArgs=0; maxArgs=0 ;;
	'build-instructions') action="${arg}"; minArgs=0; maxArgs=0 ;;
	'index-repository') action="${arg}"; minArgs=0; maxArgs=0 ;;
	'archive-contents') action="${arg}"; minArgs=0; maxArgs=0 ;;
	'prune') action="${arg}"; minArgs=1; maxArgs=1 ;;
	'repository-contents') action="${arg}"; minArgs=0; maxArgs=0 ;;
	'select') action="${arg}"; minArgs=2; maxArgs=2 ;;
	'pin') action="${arg}"; minArgs=2; maxArgs=3 ;;
	'update-image') action="${arg}"; minArgs=0; maxArgs=0 ;;
	'targets') action="${arg}"; minArgs=0; maxArgs=0 ;;
	'inside-container') action="${arg}"; minArgs=0; maxArgs=999 ;;
	*)
		reportArgumentsError "${arg}: unknown action"
		exit "${RC_BAD_ARGS}"
		;;
	esac
done

# Early exit if no action was chosen.
test -z "${action}" && { reportArgumentsError 'no action specified'; exit "${RC_BAD_ARGS}"; }

# Early exit if the number of parameters is incorrect.
test "${minArgs}" -gt 0 && test -z "$1" && { reportArgumentsError "${action}: insufficient parameters"; exit "${RC_BAD_ARGS}"; }
test "${minArgs}" -gt 1 && test -z "$2" && { reportArgumentsError "${action}: insufficient parameters"; exit "${RC_BAD_ARGS}"; }
test "${minArgs}" -eq 0 && test "${maxArgs}" -eq 0 && test -n "$1" && { reportArgumentsError "${action}: too many parameters"; exit "${RC_BAD_ARGS}"; }
test "${maxArgs}" -eq 1 && test -n "$2" && { reportArgumentsError "${action}: too many parameters"; exit "${RC_BAD_ARGS}"; }
test "${maxArgs}" -eq 2 && test -n "$3" && { reportArgumentsError "${action}: too many parameters"; exit "${RC_BAD_ARGS}"; }

# For the build-instructions action, force a target OS.
test "${action}" = 'build-instructions' && targetOs="${instructionsTarget}"

# If a target OS was selected, validate it, and set targetOsFunctionName and
# buildImageName.
test -n "${targetOs}" && setTargetOs "${targetOs}"

# Validate that the provided paths exist.
if test -n "${sourcePath}" && ! test -e "${sourcePath}"; then
	reportArgumentsError "--source: ${sourcePath}: not found"
	exit "${RC_BAD_ARGS}"
fi
if test -n "${instructionsPath}" && ! test -e "${instructionsPath}"; then
	reportArgumentsError "--instructions: ${instructionsPath}: not found"
	exit "${RC_BAD_ARGS}"
fi
if test -n "${destinationPath}" && ! test -e "${destinationPath}"; then
	reportArgumentsError "--destination: ${destinationPath}: not found"
	exit "${RC_BAD_ARGS}"
fi
if test -n "${keyFile}" && ! test -e "${keyFile}"; then
	reportArgumentsError "--key: ${keyFile}: not found"
	exit "${RC_BAD_ARGS}"
fi

# If we aren't running inside a container, then create the working
# directory, and set up an exit trap so it is removed on exit.
if ! test "${action}" = 'inside-container'; then
	workDir="$(mktemp -d)" || { reportError 'failed to create temporary working directory'; exit "${RC_LOCAL_FAULT}"; }
	trap 'rm -rf "${workDir}"' EXIT
	# Also copy the signing key into the working directory, in case it's
	# needed.
	test -n "${keyFile}" && cp "${keyFile}" "${workDir}/signing-key"
fi

# Run the appropriate action.
exitStatus="${RC_OK}"
case "${action}" in
'help')
	loadComponent 'help' || exit $?
	showHelp
	;;

'version')
	loadComponent 'version' || exit $?
	showVersion
	;;

'build-package')
	test -n "${targetOs}" || { reportArgumentsError "${action}: no target operating system specified"; exit "${RC_BAD_ARGS}"; }
	test -n "${sourcePath}" || { reportArgumentsError "${action}: no source specified"; exit "${RC_BAD_ARGS}"; }
	test -n "${destinationPath}" || { reportArgumentsError "${action}: no destination specified"; exit "${RC_BAD_ARGS}"; }

	# Check the destination target OS matches our current one.
	destinationTarget="$(cat "${destinationPath}/.target" 2>/dev/null)"
	if test -n "${destinationTarget}" && ! test "${destinationTarget}" = "${targetOs}"; then
		reportError "${destinationPath}: destination (${destinationTarget}) does not match target operating system (${targetOs})"
		exit "${RC_BAD_ARGS}"
	fi

	# First make sure the build image is present, if using containers.
	if ! "${nativeOnly}"; then
		loadComponent 'containers' || exit $?
		buildImage || exit $?
	fi

	# Extract the source.
	test -e "${workDir}/build-source" && rm -rf "${workDir}/build-source"
	mkdir -p "${workDir}/build-source"
	extractSource "${workDir}/build-source" || exit $?

	# Load the build component for the target OS.
	loadOsSpecificComponent 'build-package' || exit $?
	packageBuildFunction="${osSpecificFunction}"

	# Check that the prerequisite commands are available.
	if "${nativeOnly}"; then
		"${packageBuildFunction}" check-prerequisites || exit $?
	fi

	# Place build instructions in ${workDir}/build-instructions/.
	rm -rf "${workDir}/build-instructions"
	if test -n "${instructionsPath}"; then
		# Instructions were provided - copy them.
		if test -f "${instructionsPath}"; then
			mkdir "${workDir}/build-instructions"
			cp "${instructionsPath}" "${workDir}/build-instructions/rpm.spec"
		else
			cp -a "${instructionsPath}/" "${workDir}/build-instructions"
		fi
	else
		# Attempt to extract instructions from the source.
		loadComponent 'extract-instructions' || exit $?
		extractFunction="extract_instructions_${targetOs}"
		if type "${extractFunction}" >/dev/null 2>&1; then
			"${extractFunction}"
		else
			extractFunction="$(printf '%s\n' "${extractFunction}" | sed 's/[0-9]*$//')"
			type "${extractFunction}" >/dev/null 2>&1 && "${extractFunction}"
		fi
		# If extraction failed, the directory will be empty, so it
		# will be removable with rmdir.
		rmdir "${workDir}/build-instructions" >/dev/null 2>&1
		if ! test -d "${workDir}/build-instructions"; then
			# There were no build instructions in the source, so
			# generate some.
			loadComponent 'build-instructions' || exit $?
			mkdir "${workDir}/build-instructions"
			generateBuildInstructions "${workDir}/build-instructions" || exit $?
		fi
	fi

	# Now build the package.
	"${packageBuildFunction}"
	exitStatus="$?"

	# Record the target OS for this package archive.
	if test "${exitStatus}" -eq 0 && ! test -s "${destinationPath}/.target"; then
		printf '%s\n' "${targetOs}" > "${destinationPath}/.target"
	fi
	;;

'build-instructions')
	test -n "${sourcePath}" || { reportArgumentsError "${action}: no source specified"; exit "${RC_BAD_ARGS}"; }
	test -n "${destinationPath}" || { reportArgumentsError "${action}: no destination specified"; exit "${RC_BAD_ARGS}"; }
	loadComponent 'build-instructions' || exit $?
	generateBuildInstructions "${destinationPath}"
	exitStatus="$?"
	;;

'index-repository')
	test -n "${sourcePath}" || { reportArgumentsError "${action}: no source specified"; exit "${RC_BAD_ARGS}"; }
	test -n "${destinationPath}" || destinationPath="${sourcePath}"

	# If no target operating system was specified, see if we can read
	# one from the target path, or the source path.
	test -z "${targetOs}" && test -s "${destinationPath}/.target" && targetOs="$(cat "${destinationPath}/.target" 2>/dev/null)"
	test -z "${targetOs}" && test -s "${sourcePath}/.target" && targetOs="$(cat "${sourcePath}/.target" 2>/dev/null)"
	test -n "${targetOs}" || { reportArgumentsError "${action}: no target operating system specified"; exit "${RC_BAD_ARGS}"; }

	# Check that the prerequisite commands are available.
	if "${nativeOnly}"; then
		indexRepository "${targetOs}" "${sourcePath}" "${destinationPath}" check-prerequisites || exit $?
	fi

	indexRepository "${targetOs}" "${sourcePath}" "${destinationPath}"
	exitStatus="$?"
	;;

'archive-contents')
	test -n "${sourcePath}" || { reportArgumentsError "${action}: no source specified"; exit "${RC_BAD_ARGS}"; }
	loadComponent 'archive-contents' || exit $?
	listArchiveContents
	exitStatus="$?"
	;;

'prune')
	test -n "${sourcePath}" || { reportArgumentsError "${action}: no source specified"; exit "${RC_BAD_ARGS}"; }
	pruneCount="$(printf '%s\n' "$1" | tr -dc '0-9')"
	test -n "${pruneCount}" || { reportArgumentsError "${action}: $1: number not recognised"; exit "${RC_BAD_ARGS}"; }
	loadComponent 'prune-archive' || exit $?
	pruneArchive
	exitStatus="$?"
	;;

'repository-contents')
	test -n "${destinationPath}" || { reportArgumentsError "${action}: no destination specified"; exit "${RC_BAD_ARGS}"; }
	loadComponent 'repository-contents' || exit $?
	listRepositoryCollectionContents
	exitStatus="$?"
	;;

'select')
	test -n "${sourcePath}" || { reportArgumentsError "${action}: no source specified"; exit "${RC_BAD_ARGS}"; }
	test -n "${destinationPath}" || { reportArgumentsError "${action}: no destination specified"; exit "${RC_BAD_ARGS}"; }
	loadComponent 'select' || exit $?
	selectPackage "$1" "$2" 'false'
	exitStatus="$?"
	;;

'pin')
	test -n "${sourcePath}" || { reportArgumentsError "${action}: no source specified"; exit "${RC_BAD_ARGS}"; }
	test -n "${destinationPath}" || { reportArgumentsError "${action}: no destination specified"; exit "${RC_BAD_ARGS}"; }
	loadComponent 'select' || exit $?
	selectPackage "$1" "$2" 'true' "$3"
	exitStatus="$?"
	;;

'update-image')
	test -n "${targetOs}" || { reportArgumentsError "${action}: no target operating system specified"; exit "${RC_BAD_ARGS}"; }
	"${nativeOnly}" && { reportArgumentsError "${action}: not compatible with --native"; exit "${RC_BAD_ARGS}"; }
	loadComponent 'containers' || exit $?
	updateImage
	exitStatus="$?"
	;;

'targets')
	printf '%s\n' "${supportedTargets}" | tr ' ' '\n' | grep -E .
	;;

'inside-container')
	runFromInsideContainer "$@"
	;;

esac

# TODO: option to defer reindexing when calling "select" and "pin"
# TODO: action to run deferred reindexing
# TODO: allow select/pin to fuzzy-match on version without "-release" component

exit "${exitStatus}"
